/***************************************************************************

  M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
  Win32 Portions Copyright (C) 1997-2001 Michael Soderstrom and Chris Kirmse

  This file is part of MAME32, and may only be used, modified and
  distributed under the terms of the MAME license, in "readme.txt".
  By continuing to use, modify or distribute this file you indicate
  that you have read the license and understand and accept it fully.

 ***************************************************************************/
 
 /***************************************************************************

  options.c

  Stores global options and per-game options;

***************************************************************************/

#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <windowsx.h>
#include <libintl.h>
#include <commctrl.h>
#include <assert.h>
#include <stdio.h>
#include <io.h>
#include <sys/stat.h>
#include <malloc.h>
#include <math.h>
#include "driver.h"
#include "mame32.h"
#include "m32util.h"
#include "resource.h"
#include "windows/rc.h"

/***************************************************************************
    Internal function prototypes
 ***************************************************************************/

extern void  alloclog (char *fmt, ...);
extern void  report_allocated_block (void);

static int   rc_load_config (int game_index);
static int   rc_save_config (int game_index);

static int   initialize_rc_winui_config (void);
static int   rc_load_winui_config (void);
static int   rc_save_winui_config (void);
static int   rc_write_folder_flags (FILE *fp);

static void  backup_settings (settings_type *p);
static void  free_settings (settings_type *p);

static void  free_if_allocated (char **p);

static void  ValidateDefaultOptions (void);

static void  SaveOptions (void);
static void  SaveGlobalOptions (void);
static void  LoadOptions (void);
static void  LoadGameOptions (int game_index);

static char  *ColumnEncodeString (int *data);
static int   ColumnDecodeString (struct rc_option *option, const char *str, int* data);

static void  ColumnOrderEncodeString (void);
static int   ColumnOrderDecodeString (struct rc_option *option, const char *arg, int priority);

static void  ColumnShownEncodeString (void);
static int   ColumnShownDecodeString (struct rc_option *option, const char *arg, int priority);

static void  ColumnEncodeWidths (void);
static int   ColumnDecodeWidths (struct rc_option *option, const char *arg, int priority);

static void  SplitterEncodeString (void);
static int   SplitterDecodeString (struct rc_option *option, const char *arg, int priority);

static void  ListEncodeString (void);
static int   ListDecodeString (struct rc_option *option, const char *arg, int priority);

static void  FontEncodeString (void);
static int   FontDecodeString (struct rc_option *option, const char *arg, int priority);

static int   FontDecodeColor (struct rc_option *option, const char *arg, int priority);

static void  SaveFolderFlags (char *folderName, DWORD dwFlags);

static int   FolderFlagDecodeString (struct rc_option *option, const char *arg, int priority);

/***************************************************************************
    Internal defines
 ***************************************************************************/

#define FOLDERFLAG_OPT		"folder_flag"
#define ALLOC_FOLDERFLAG	8

/***************************************************************************
    Internal structures
 ***************************************************************************/

/***************************************************************************
    Internal variables
 ***************************************************************************/

static settings_type settings;

static struct {
	settings_type settings;
	options_type  global;
} backup;

static options_type gOpts;  /* Used when saving/loading from rc file */
static options_type global; /* Global 'default' options */
static options_type *game;  /* Array of Game specific options */

static int num_folders = 0;
static struct folder_flags_type {
	char  name[80];
	DWORD flags;
} *folder_flags;

/* Global UI options */
static int  num_games = 0;
static BOOL bResetGUI      = FALSE;

static char *view_modes[VIEW_MAX] = { "Large Icons", "Small Icons", "List", "Details" };

static char oldInfoMsg[400] = 
MAME32NAME " has detected outdated configuration data.\n\n\
The detected configuration data is from Version %s of " MAME32NAME ".\n\
The current version is %s. It is recommended that the\n\
configuration is set to the new defaults.\n\n\
Would you like to use the new configuration?";

static char reload_config_msg[] =
MAME32NAME " has changed *.ini file directory.\n\n\
Would you like to use the configurations in this directory?";

//============================================================
//	rc options
//============================================================

static struct rc_struct *rc_core;
static struct rc_struct *rc_game;
static struct rc_struct **rc_winui;

/* temporary for rc */
static struct {
	/* MAMEW core */
	char *playbackname;
	char *recordname;

	/* WINUI */
	char save_version[80];
	char view[16];
	char list_font[80];
	char splitter[80];
	char column_width[80];
	char column_order[80];
	char column_shown[80];
} rc_dummy_args;

static struct rc_option rc_game_opts[] =
{
	{ "Windows video options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "autoframeskip", "afs", rc_bool, &gOpts.autoframeskip, "1", 0, 0, NULL, "skip frames to speed up emulation" },
	{ "frameskip", "fs", rc_int, &gOpts.frameskip, "0", 0, 12, NULL, "set frameskip explicitly (autoframeskip needs to be off)" },
	{ "waitvsync", NULL, rc_bool, &gOpts.wait_vsync, "0", 0, 0, NULL, "wait for vertical sync (reduces tearing)"},
	{ "triplebuffer", "tb", rc_bool, &gOpts.use_triplebuf, "0", 0, 0, NULL, "triple buffering (only if fullscreen)" },
	{ "window", "w", rc_bool, &gOpts.window_mode, "0", 0, 0, NULL, "run in a window/run on full screen" },
	{ "ddraw", "dd", rc_bool, &gOpts.use_ddraw, "1", 0, 0, NULL, "use DirectDraw for rendering" },
	{ "hwstretch", "hws", rc_bool, &gOpts.ddraw_stretch, "1", 0, 0, NULL, "stretch video using the hardware" },
	{ "resolution", "r", rc_sstring, &gOpts.resolution, "auto", 0, 0, NULL, "set resolution" },
	{ "refresh", NULL, rc_int, &gOpts.gfx_refresh, "0", 0, 0, NULL, "set specific monitor refresh rate" },
	{ "scanlines", "sl", rc_bool, &gOpts.scanlines, "0", 0, 0, NULL, "emulate win_old_scanlines" },
	{ "switchres", NULL, rc_bool, &gOpts.switchres, "1", 0, 0, NULL, "switch resolutions to best fit" },
	{ "switchbpp", NULL, rc_bool, &gOpts.switchbpp, "1", 0, 0, NULL, "switch color depths to best fit" },
	{ "maximize", "max", rc_bool, &gOpts.maximize, "1", 0, 0, NULL, "start out maximized" },
	{ "keepaspect", "ka", rc_bool, &gOpts.keepaspect, "1", 0, 0, NULL, "enforce aspect ratio" },
	{ "matchrefresh", NULL, rc_bool, &gOpts.matchrefresh, "0", 0, 0, NULL, "attempt to match the game's refresh rate" },
	{ "syncrefresh", NULL, rc_bool, &gOpts.syncrefresh, "0", 0, 0, NULL, "syncronize only to the monitor refresh" },
	{ "dirty", NULL, rc_bool, &gOpts.use_dirty, "1", 0, 0, NULL, "enable dirty video optimization" },
	{ "throttle", NULL, rc_bool, &gOpts.throttle, "1", 0, 0, NULL, "throttle speed to the game's framerate" },
	{ "full_screen_brightness", "fsb", rc_float, &gOpts.gfx_brightness, "0.0", 0.0, 4.0, NULL, "sets the brightness in full screen mode" },
	{ "frames_to_run", "ftr", rc_int, &gOpts.frames_to_display, "0", 0, 0, NULL, "sets the number of frames to run within the game" },
	{ "effect", NULL, rc_sstring, &gOpts.effect, "none", 0, 0, NULL, "specify the blitting effect" },
#ifdef USE_2XSAIMODE
	{ "saimode", NULL, rc_int, &gOpts.saimode, "0", 0, 4, NULL, "specify the 2xsai mode" },
#endif /* USE_2XSAIMODE */
	{ "screen_aspect", NULL, rc_sstring, &gOpts.aspect, "4:3", 0, 0, NULL, "specify an alternate monitor aspect ratio" },
	{ "sleep", NULL, rc_bool, &gOpts.allow_sleep, "1", 0, 0, NULL, "allow MAME to give back time to the system when it's not needed" },
#ifdef USE_DIRECT3D_EFFECT
	{ "spurious-motion-blur", "smb", rc_bool, &gOpts.motionblur, "0", 0, 0, NULL, "spurious motion blur" },
	{ "bilinear-filter", "blf", rc_bool, &gOpts.bilinear, "1", 0, 0, NULL, "stretch with bilinear filter" },
#endif /* USE_DIRECT3D_EFFECT */

	{ "Input device options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "hotrod", NULL, rc_bool, &gOpts.hotrod, "0", 0, 0, NULL, "preconfigure for hotrod" },
	{ "hotrodse", NULL, rc_bool, &gOpts.hotrodse, "0", 0, 0, NULL, "preconfigure for hotrod se" },
	{ "mouse", NULL, rc_bool, &gOpts.use_mouse, "0", 0, 0, NULL, "enable mouse input" },
	{ "joystick", "joy", rc_bool, &gOpts.use_joystick, "0", 0, 0, NULL, "enable joystick input" },
	{ "steadykey", "steady", rc_bool, &gOpts.steadykey, "0", 0, 0, NULL, "enable steadykey support" },
	{ "ctrlr", NULL, rc_sstring, &gOpts.ctrlr, "Standard", 0, 0, NULL, "specify the controller definitions" },

	/* options supported by the mame core */
	/* video */
	{ "Mame CORE video options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "bpp", NULL, rc_int, &gOpts.color_depth, "0",	0, 0, NULL, "specify the colordepth the core should render in bits per pixel (bpp), one of: auto (0), 8, 16, 32" },
	{ "norotate", NULL, rc_bool , &gOpts.norotate, "0", 0, 0, NULL, "do not apply rotation" },
	{ "ror", NULL, rc_bool, &gOpts.ror, "0", 0, 0, NULL, "rotate screen clockwise" },
	{ "rol", NULL, rc_bool, &gOpts.rol, "0", 0, 0, NULL, "rotate screen anti-clockwise" },
	{ "flipx", NULL, rc_bool, &gOpts.flipx, "0", 0, 0, NULL, "flip screen upside-down" },
	{ "flipy", NULL, rc_bool, &gOpts.flipy, "0", 0, 0, NULL, "flip screen left-right" },
	{ "debug_resolution", "dr", rc_sstring, &gOpts.debugres, "auto", 0, 0, NULL, "set resolution for debugger window" },
	/* make it gOpts.gamma_correction? */
	{ "gamma", NULL, rc_float, &gOpts.gamma_correct , "1.0", 0.5, 2.0, NULL, "gamma correction"},

	/* vector */
	{ "Mame CORE vector game options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "antialias", "aa", rc_bool, &gOpts.antialias, "1", 0, 0, NULL, "draw antialiased vectors" },
	{ "translucency", "tl", rc_bool, &gOpts.translucency, "1", 0, 0, NULL, "draw translucent vectors" },
	{ "beam", NULL, rc_float, &gOpts.f_beam, "1.0", 1.0, 16.0, NULL, "set beam width in vector games" },
	{ "flicker", NULL, rc_float, &gOpts.f_flicker, "0.0", 0.0, 100.0, NULL, "set flickering in vector games" },

	/* sound */
	{ "Mame CORE sound options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "samplerate", "sr", rc_int, &gOpts.samplerate, "44100", 5000, 50000, NULL, "set samplerate" },
	{ "samples", NULL, rc_bool, &gOpts.use_samples, "1", 0, 0, NULL, "use samples" },
	{ "resamplefilter", NULL, rc_bool, &gOpts.use_filter, "1", 0, 0, NULL, "resample if samplerate does not match" },
	{ "sound", NULL, rc_bool, &gOpts.enable_sound, "1", 0, 0, NULL, "enable/disable sound and sound CPUs" },
	{ "volume", "vol", rc_int, &gOpts.attenuation, "0", -32, 0, NULL, "volume (range [-32,0])" },

	/* misc */
	{ "Mame CORE misc options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "artwork", "art", rc_bool, &gOpts.use_artwork, "1", 0, 0, NULL, "use additional game artwork" },
	{ "cheat", "c", rc_bool, &gOpts.cheat, "1", 0, 0, NULL, "enable/disable cheat subsystem" },
	{ "debug", "d", rc_bool, &gOpts.mame_debug, "0", 0, 0, NULL, "enable/disable debugger (only if available)" },
	{ "playback", "pb", rc_string, &rc_dummy_args.playbackname, NULL, 0, 0, NULL, "playback an input file" },
	{ "record", "rec", rc_string, &rc_dummy_args.recordname, NULL, 0, 0, NULL, "record an input file" },
	{ "log", NULL, rc_bool, &gOpts.errorlog, "0", 0, 0, NULL, "generate error.log" },

	{ NULL,	NULL, rc_end, NULL, NULL, 0, 0,	NULL, NULL }
};

static struct rc_option rc_mamew_opts[] =
{
	/* name, shortname, type, dest, deflt, min, max, func, help */
	{ "Frontend Related", NULL,	rc_seperator, NULL, NULL, 0, 0,	NULL, NULL },
	{ "clones", NULL, rc_bool, &gOpts.listclones, "1", 0, 0, NULL, "enable/disable clones" },

	{ "Windows path and directory options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "rompath", "rp", rc_string, &settings.romdirs, "roms", 0, 0, NULL, "path to romsets" },
	{ "samplepath", "sp", rc_string, &settings.sampledirs, "samples", 0, 0, NULL, "path to samplesets" },
	{ "cfg_directory", NULL, rc_string, &settings.cfgdir, "cfg", 0, 0, NULL, "directory to save configurations" },
	{ "nvram_directory", NULL, rc_string, &settings.nvramdir, "nvram", 0, 0, NULL, "directory to save nvram contents" },
	{ "memcard_directory", NULL, rc_string, &settings.memcarddir, "memcard", 0, 0, NULL, "directory to save memory card contents" },
	{ "input_directory", NULL, rc_string, &settings.inpdir, "inp", 0, 0, NULL, "directory to save input device logs" },
	{ "hiscore_directory", NULL, rc_string, &settings.hidir, "hi", 0, 0, NULL, "directory to save hiscores" },
	{ "state_directory", NULL, rc_string, &settings.statedir, "sta", 0, 0, NULL, "directory to save states" },
	{ "artwork_directory", NULL, rc_string, &settings.artdir, "artwork", 0, 0, NULL, "directory for Artwork (Overlays etc.)" },
	{ "snapshot_directory", NULL, rc_string, &settings.imgdir, "snap", 0, 0, NULL, "directory for screenshots (.png format)" },
	{ "diff_directory", NULL, rc_string, &settings.diffdir, "diff", 0, 0, NULL, "directory for hard drive image difference files" },
	{ "cheat_file", NULL, rc_string, &settings.cheatfile, "cheat.dat", 0, 0, NULL, "cheat filename" },
	{ "history_file", NULL, rc_string, &settings.history_filename, "history.dat", 0, 0, NULL, NULL },
	{ "mameinfo_file", NULL, rc_string, &settings.mameinfo_filename, "mameinfo.dat", 0, 0, NULL, NULL },
	{ "ctrlr_directory", NULL, rc_string, &settings.ctrlrdir, "ctrlr", 0, 0, NULL, "directory to save controller definitions" },

	{ NULL, NULL, rc_link, rc_game_opts, NULL, 0,	0, NULL, NULL },

	{ "Configuration options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "verbose", "v", rc_bool, &gOpts.verbose, "0", 0, 0, NULL, "display additional diagnostic information" },

	{ "Language options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "use_lang_list", NULL, rc_bool, &settings.use_lang_list, "1", 0, 0, NULL, "enable/disable language list" },
	{ NULL,	NULL, rc_end, NULL, NULL, 0, 0,	NULL, NULL }
};

static struct rc_option rc_winui_opts[] =
{
	{ "WINUI specific path options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{ "language", NULL, rc_string, &settings.language, "english", 0, 0, NULL, "select mame ui language" },
	{ "flyer_directory", NULL, rc_string, &settings.flyerdir, "flyers", 0, 0, NULL, "directory for flyers" },
	{ "cabinet_directory", NULL, rc_string, &settings.cabinetdir, "cabinets", 0, 0, NULL, "directory for cabinets" },
	{ "marquee_directory", NULL, rc_string, &settings.marqueedir, "marquees", 0, 0, NULL, "directory for marquees" },
#ifdef PLUS
	{ "cpanel_directory", NULL, rc_string, &settings.cpaneldir, "cpanel", 0, 0, NULL, "directory for cpanel" },
	{ "title_directory", NULL, rc_string, &settings.titledir, "titles", 0, 0, NULL, "directory for titles" },
#endif /* PLUS */
	{ "ini_directory", NULL, rc_string, &settings.inidir, "ini", 0, 0, NULL, "directory for ini files" },
	{ "icons_directory", NULL, rc_string, &settings.iconsdir, "icons", 0, 0, NULL, "directory for icons" },
#ifdef EXTRA_FOLDER
	{ "folder_directory", NULL, rc_string, &settings.folderdir, "folders", 0, 0, NULL, "directory for folders" },
#endif /* EXTRA_FOLDER */

	{ "WINUI specific startup options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{"reset_gui", NULL, rc_bool, &bResetGUI, "0", 0, 0, NULL, "enable version mismatch warning" },
	{"version_check", NULL, rc_bool, &settings.version_check, "1", 0, 0, NULL, "enable version mismatch warning" },
	{"save_version", NULL, rc_sstring, rc_dummy_args.save_version, "", 0, 0, NULL, "config version" },
	{"game_check", NULL, rc_bool, &settings.game_check, "1", 0, 0, NULL, "search for new games" },
	{"use_joygui", NULL, rc_bool, &settings.use_joygui, "0", 0, 0, NULL, "allow game selection by a joystick" },
	{"broadcast", NULL, rc_bool, &settings.broadcast, "0", 0, 0, NULL, "broadcast selected game to all windows" },

	{ "WINUI specific general options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{"default_game", NULL, rc_sstring, settings.default_game, "pacman", 0, 0, NULL, "last selected game name" },
	{"show_screenshot", NULL, rc_bool, &settings.show_screenshot, "1", 0, 0, NULL, "show image picture" },
	{"show_pict_type", NULL, rc_int, &settings.show_pict_type, "0", 0, MAX_PICT_TYPES-1, NULL, "select image picture type" },
	{"show_toolbar", NULL, rc_bool, &settings.show_toolbar, "1", 0, 0, NULL, "show tool bar" },
	{"show_statusbar", NULL, rc_bool, &settings.show_statusbar, "1", 0, 0, NULL, "show status bar" },
	{"show_folderlist", NULL, rc_bool, &settings.show_folderlist, "1", 0, 0, NULL, "show folder list" },

	{ "WINUI specific windows position options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{"window_left", NULL, rc_int, &settings.area.x, "0", 0, (UINT)-1, NULL, "window left position" },
	{"window_top", NULL, rc_int, &settings.area.y, "0", 0, (UINT)-1, NULL, "window top position" },
	{"window_width", NULL, rc_int, &settings.area.width, "640", 550, (UINT)-1, NULL, "window width" },
	{"window_height", NULL, rc_int, &settings.area.height, "400", 400, (UINT)-1, NULL, "window height" },

	{ "WINUI specific list options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{"list_mode", NULL, rc_sstring, rc_dummy_args.view, "Details", 0, 0, ListDecodeString, "view mode" },
	{"splitter", NULL, rc_sstring, rc_dummy_args.splitter, "150,300", 0, 0, SplitterDecodeString, "splitter position" },
/* re-arrange default column_width, column_order, sort_column */
	{"column_width", NULL, rc_sstring, rc_dummy_args.column_width, "186,84,84,108,60,84,144,84,84,84,84", 0, 0, ColumnDecodeWidths, "column width settings" },
	{"column_order", NULL, rc_sstring, rc_dummy_args.column_order, "0,2,3,4,5,6,7,8,9,10,1", 0, 0, ColumnOrderDecodeString, "column order settings" },
	{"column_shown", NULL, rc_sstring, rc_dummy_args.column_shown, "1,0,1,1,1,1,1,1,1,1,1", 0, 0, ColumnShownDecodeString, "shown/hide column settings" },
	{"folder_id", NULL, rc_int, &settings.folder_id, "0", 0, (UINT)-1, NULL, "last selected folder id" },
	{"sort_column", NULL, rc_int, &settings.sort_column, "2", 0, COLUMN_MAX-1, NULL, "sort column" },
	{"sort_reverse", NULL, rc_bool, &settings.sort_reverse, "0", 0, 0, NULL, "Sort descending" },

	{ "WINUI specific list font options", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
	{"list_font", NULL, rc_sstring, rc_dummy_args.list_font, "-8,0,0,0,400,0,0,0,0,0,0,0,0", 0, 0, FontDecodeString, "list font settings" },
	{"list_font_face", NULL, rc_sstring, settings.list_font.lfFaceName, "MS Sans Serif", 0, 0, NULL, "list font face" },
	{"list_font_color", NULL, rc_int, &settings.list_font_color, "-1", -1, (UINT)-1, FontDecodeColor, "list font color" },

	{FOLDERFLAG_OPT, NULL, rc_use_function, NULL, NULL, 0, 0, FolderFlagDecodeString, "folder flag settings" },

	{ NULL,	NULL, rc_end, NULL, NULL, 0, 0,	NULL, NULL }
};

/***************************************************************************
    External functions  
 ***************************************************************************/

void OptionsInit (int total_games)
{
	int i;

	num_games = total_games;

	if (! (rc_core = rc_create ()))
		exit (1);

	if (rc_register (rc_core, rc_mamew_opts))
		exit (1);

	if (! (rc_game = rc_create ()))
		exit (1);

	if (rc_register (rc_game, rc_game_opts))
		exit (1);

	global = gOpts;
	global.use_default	= TRUE;
	global.is_favorite	= FALSE;
	global.play_count	= 0;
	global.has_roms		= UNKNOWN;
	global.has_samples	= UNKNOWN;

	/* This allocation should be checked */
	game = (options_type *)malloc (num_games * sizeof (options_type));

	for (i = 0; i < num_games; i++)
		game[i] = global;

	initialize_rc_winui_config ();

	// Create Backup
	backup.settings = settings;
	backup.global = global;
	backup_settings (&backup.settings);

	ValidateDefaultOptions ();
	LoadOptions ();
}

void OptionsExit (void)
{
	SaveOptions ();

	if (folder_flags)
		free (folder_flags);

	free (game);
	free_settings (&settings);

}

options_type * GetDefaultOptions (void)
{
	return &global;
}

options_type * GetGameOptions (int num_game)
{
	int play_count;
	int has_roms;
	int has_samples;
	BOOL is_favorite;

	assert (0 <= num_game && num_game < num_games);

	play_count	= game[num_game].play_count;
	has_roms	= game[num_game].has_roms;
	has_samples	= game[num_game].has_samples;
	is_favorite	= game[num_game].is_favorite;

	if (game[num_game].use_default)
	{
		game[num_game]			= global;
		game[num_game].use_default	= TRUE;
		game[num_game].play_count	= play_count;
		game[num_game].has_roms 	= has_roms;
		game[num_game].has_samples	= has_samples;
		game[num_game].is_favorite	= is_favorite;

	}
	return &game[num_game];
}

void ResetGUI (void)
{
	bResetGUI = TRUE;
}

BOOL UseLangList(void)
{
    return settings.use_lang_list;
}

void SetUseLangList(BOOL is_use)
{
    settings.use_lang_list = is_use;
}

void SetViewMode (int val)
{
	settings.view = val;
}

int GetViewMode (void)
{
	return settings.view;
}

void SetGameCheck (BOOL game_check)
{
	settings.game_check = game_check;
}

BOOL GetGameCheck (void)
{
	return settings.game_check;
}

void SetVersionCheck (BOOL version_check)
{
	settings.version_check = version_check;
}

BOOL GetVersionCheck (void)
{
	return settings.version_check;
}

void SetJoyGUI (BOOL use_joygui)
{
	settings.use_joygui = use_joygui;
}

BOOL GetJoyGUI (void)
{
	return settings.use_joygui;
}

void SetBroadcast (BOOL broadcast)
{
	settings.broadcast = broadcast;
}

BOOL GetBroadcast (void)
{
	return settings.broadcast;
}

void SetSavedFolderID (UINT val)
{
	settings.folder_id = val;
}

UINT GetSavedFolderID (void)
{
	return settings.folder_id;
}

void SetShowScreenShot (BOOL val)
{
	settings.show_screenshot = val;
}

BOOL GetShowScreenShot (void)
{
	return settings.show_screenshot;
}

void SetShowFolderList (BOOL val)
{
	settings.show_folderlist = val;
}

BOOL GetShowFolderList (void)
{
	return settings.show_folderlist;
}

void SetShowStatusBar (BOOL val)
{
	settings.show_statusbar = val;
}

BOOL GetShowStatusBar (void)
{
	return settings.show_statusbar;
}

void SetShowToolBar (BOOL val)
{
	settings.show_toolbar = val;
}

BOOL GetShowToolBar (void)
{
	return settings.show_toolbar;
}

void SetShowPictType (int val)
{
	settings.show_pict_type = val;
}

int GetShowPictType (void)
{
	return settings.show_pict_type;
}

void SetDefaultGame (const char *name)
{
	strcpy (settings.default_game, name);
}

const char *GetDefaultGame (void)
{
	return settings.default_game;
}

void SetWindowArea (AREA *area)
{
	memcpy (&settings.area, area, sizeof (AREA));
}

void GetWindowArea (AREA *area)
{
	memcpy (area, &settings.area, sizeof (AREA));
}

void SetListFont (LOGFONT *font)
{
	memcpy (&settings.list_font, font, sizeof (LOGFONT));
}

void GetListFont (LOGFONT *font)
{
	memcpy (font, &settings.list_font, sizeof (LOGFONT));
}

void SetListFontColor (COLORREF uColor)
{
	if (settings.list_font_color == GetSysColor (COLOR_WINDOWTEXT))
		settings.list_font_color = (COLORREF)-1;
	else
		settings.list_font_color = uColor;
}

COLORREF GetListFontColor (void)
{
	if (settings.list_font_color == (COLORREF)-1)
		return (GetSysColor (COLOR_WINDOWTEXT));

	return settings.list_font_color;
}

void SetColumnWidths (int width[])
{
	int i;

	for (i = 0; i < COLUMN_MAX; i++)
		settings.column_width[i] = width[i];
}

void GetColumnWidths (int width[])
{
	int i;

	for (i = 0; i < COLUMN_MAX; i++)
		width[i] = settings.column_width[i];
}

void SetSplitterPos (int splitterId, int pos)
{
	if (splitterId < SPLITTER_MAX)
		settings.splitter[splitterId] = pos;
}

int  GetSplitterPos (int splitterId)
{
	if (splitterId < SPLITTER_MAX)
		return settings.splitter[splitterId];

	return -1; /* Error */
}

void SetColumnOrder (int order[])
{
	int i;

	for (i = 0; i < COLUMN_MAX; i++)
		settings.column_order[i] = order[i];
}

void GetColumnOrder (int order[])
{
	int i;

	for (i = 0; i < COLUMN_MAX; i++)
		order[i] = settings.column_order[i];
}

void SetColumnShown (int shown[])
{
	int i;

	for (i = 0; i < COLUMN_MAX; i++)
		settings.column_shown[i] = shown[i];
}

void GetColumnShown (int shown[])
{
	int i;

	for (i = 0; i < COLUMN_MAX; i++)
		shown[i] = settings.column_shown[i];
}

void SetSortColumn (int column)
{
	settings.sort_reverse = (column < 0) ? TRUE : FALSE;
	settings.sort_column  = abs (column) - 1;
}

int GetSortColumn (void)
{
	int column = settings.sort_column + 1;

	if (settings.sort_reverse)
		column = - (column);

	return column;
}

const char* GetLanguage (void)
{
	return settings.language;
}

void SetLanguage (const char* lang)
{
	free_if_allocated (&settings.language);

	if (lang != NULL)
		settings.language = strdup (lang);
}

#ifdef EXTRA_FOLDER
const char *GetFolderDir(void)
{
	return settings.folderdir;
}

void SetFolderDir(const char *path)
{
	free_if_allocated (&settings.folderdir);

	if (path != NULL)
		settings.folderdir = strdup(path);
}
#endif /* EXTRA_FOLDER */

const char* GetIniDir (void)
{
	return settings.inidir;
}

void SetIniDir (const char* path)
{
	if (!strcmp (path, settings.inidir))
		return;

	free_if_allocated (&settings.inidir);

	if (path != NULL)
		settings.inidir = strdup (path);

	if (MessageBox (0, reload_config_msg, "Reload configurations", MB_YESNO | MB_ICONQUESTION) == IDYES)
	{
		int i;

		for (i = 0 ; i < num_games; i++)
			LoadGameOptions (i);
	}

	_mkdir (path);
}

const char* GetIconsDir (void)
{
	return settings.iconsdir;
}

void SetIconsDir (const char* path)
{
	free_if_allocated (&settings.iconsdir);

	if (path != NULL)
		settings.iconsdir = strdup (path);
}

const char* GetRomDirs (void)
{
	return settings.romdirs;
}

void SetRomDirs (const char* paths)
{
	free_if_allocated (&settings.romdirs);

	if (paths != NULL)
		settings.romdirs = strdup (paths);
}

const char* GetSampleDirs (void)
{
	return settings.sampledirs;
}

void SetSampleDirs (const char* paths)
{
	free_if_allocated (&settings.sampledirs);

	if (paths != NULL)
		settings.sampledirs = strdup (paths);
}

const char* GetCtrlrDir (void)
{
	return settings.ctrlrdir;
}

void SetCtrlrDir (const char* path)
{
	free_if_allocated (&settings.ctrlrdir);

	if (path != NULL)
		settings.ctrlrdir = strdup (path);
}

const char* GetCfgDir (void)
{
	return settings.cfgdir;
}

void SetCfgDir (const char* path)
{
	free_if_allocated (&settings.cfgdir);

	if (path != NULL)
		settings.cfgdir = strdup (path);
}

const char* GetHiDir (void)
{
	return settings.hidir;
}

void SetHiDir (const char* path)
{
	free_if_allocated (&settings.hidir);

	if (path != NULL)
		settings.hidir = strdup (path);
}

const char* GetNvramDir (void)
{
	return settings.nvramdir;
}

void SetNvramDir (const char* path)
{
	free_if_allocated (&settings.nvramdir);

	if (path != NULL)
		settings.nvramdir = strdup (path);
}

const char* GetInpDir (void)
{
	return settings.inpdir;
}

void SetInpDir (const char* path)
{
	free_if_allocated (&settings.inpdir);

	if (path != NULL)
		settings.inpdir = strdup (path);
}

const char* GetImgDir (void)
{
	return settings.imgdir;
}

void SetImgDir (const char* path)
{
	free_if_allocated (&settings.imgdir);

	if (path != NULL)
		settings.imgdir = strdup (path);
}

const char* GetStateDir (void)
{
	return settings.statedir;
}

void SetStateDir (const char* path)
{
	free_if_allocated (&settings.statedir);

	if (path != NULL)
		settings.statedir = strdup (path);
}

const char* GetArtDir (void)
{
	return settings.artdir;
}

void SetArtDir (const char* path)
{
	free_if_allocated (&settings.artdir);

	if (path != NULL)
		settings.artdir = strdup (path);
}

const char* GetMemcardDir (void)
{
	return settings.memcarddir;
}

void SetMemcardDir (const char* path)
{
	free_if_allocated (&settings.memcarddir);

	if (path != NULL)
		settings.memcarddir = strdup (path);
}

const char* GetFlyerDir (void)
{
	return settings.flyerdir;
}

void SetFlyerDir (const char* path)
{
	free_if_allocated (&settings.flyerdir);

	if (path != NULL)
		settings.flyerdir = strdup (path);
}

const char* GetCabinetDir (void)
{
	return settings.cabinetdir;
}

void SetCabinetDir (const char* path)
{
	free_if_allocated (&settings.cabinetdir);

	if (path != NULL)
		settings.cabinetdir = strdup (path);
}

const char* GetMarqueeDir (void)
{
	return settings.marqueedir;
}

void SetMarqueeDir (const char* path)
{
	free_if_allocated (&settings.marqueedir);

	if (path != NULL)
		settings.marqueedir = strdup (path);
}

#ifdef PLUS
const char* GetCpanelDir (void)
{
	return settings.cpaneldir;
}

void SetCpanelDir (const char* path)
{
	free_if_allocated (&settings.cpaneldir);

	if (path != NULL)
		settings.cpaneldir = strdup (path);
}

const char* GetTitleDir (void)
{
	return settings.titledir;
}

void SetTitleDir (const char* path)
{
	free_if_allocated (&settings.titledir);

	if (path != NULL)
		settings.titledir = strdup (path);
}
#endif /* PLUS */

const char* GetDiffDir (void)
{
	return settings.diffdir;
}

void SetDiffDir (const char* path)
{
	free_if_allocated (&settings.diffdir);

	if (path != NULL)
		settings.diffdir = strdup (path);
}

const char* GetCheatFileName (void)
{
	return settings.cheatfile;
}

void SetCheatFileName (const char* path)
{
	free_if_allocated (&settings.cheatfile);

	if (path != NULL)
		settings.cheatfile = strdup (path);
}

const char* GetHistoryFileName (void)
{
	return settings.history_filename;
}

void SetHistoryFileName (const char* path)
{
	free_if_allocated (&settings.history_filename);

	if (path != NULL)
		settings.history_filename = strdup (path);
}


const char* GetMAMEInfoFileName (void)
{
	return settings.mameinfo_filename;
}

void SetMAMEInfoFileName (const char* path)
{
	free_if_allocated (&settings.mameinfo_filename);

	if (path != NULL)
		settings.mameinfo_filename = strdup (path);
}

void ResetGameOptions (int num_game)
{
	int play_count;
	int has_roms;
	int has_samples;

	assert (0 <= num_game && num_game < num_games);

	play_count =	game[num_game].play_count;
	has_roms =	game[num_game].has_roms;
	has_samples =	game[num_game].has_samples;

	game[num_game] =		global;
	game[num_game].use_default =	TRUE;
	game[num_game].play_count =	play_count;
	game[num_game].has_roms =	has_roms;
	game[num_game].has_samples =	has_samples;
}

void ResetGameDefaults (void)
{
	global = backup.global;

	ValidateDefaultOptions ();
}

void ResetAllGameOptions (void)
{
	int i;

	for (i = 0; i < num_games; i++) {
		ResetGameOptions (i);
		game[i].play_count = 0;
	}
}

int  GetHasRoms (int num_game)
{
	assert (0 <= num_game && num_game < num_games);

	return game[num_game].has_roms;
}

void SetHasRoms (int num_game, int has_roms)
{
	assert (0 <= num_game && num_game < num_games);

	game[num_game].has_roms = has_roms;
}

int  GetHasSamples (int num_game)
{
	assert (0 <= num_game && num_game < num_games);

	return game[num_game].has_samples;
}

void SetHasSamples (int num_game, int has_samples)
{
	assert (0 <= num_game && num_game < num_games);

	game[num_game].has_samples = has_samples;
}

int  GetIsFavorite (int num_game)
{
	assert (0 <= num_game && num_game < num_games);

	return game[num_game].is_favorite;
}

void SetIsFavorite (int num_game, int is_favorite)
{
	assert (0 <= num_game && num_game < num_games);

	game[num_game].is_favorite = is_favorite;
}

void IncrementPlayCount (int num_game)
{
	assert (0 <= num_game && num_game < num_games);

	game[num_game].play_count++;
}

void SetFolderFlags (char *folderName, DWORD dwFlags)
{
	SaveFolderFlags (folderName, dwFlags);
}

int GetPlayCount (int num_game)
{
	assert (0 <= num_game && num_game < num_games);

	return game[num_game].play_count;
}

char* GetVersionString (void)
{
	return build_version;
}

DWORD GetFolderFlags (char *folderName)
{
	int i;

	for (i = 0; i < num_folders; i++)
		if (!strcmp (folderName, folder_flags[i].name))
			return folder_flags[i].flags;

	return 0;
}

void SaveGameOptions (int game_index)
{
	assert (0 <= game_index && game_index < num_games);

	gOpts = *GetGameOptions (game_index);

	rc_save_config (game_index);
}

void SaveDefaultOptions (void)
{
	ValidateDefaultOptions ();

	gOpts = global;

	rc_save_config (-1);
}

void SaveOptions (void)
{
	int i;

	SaveGlobalOptions ();

	for (i = 0; i < num_games; i++)
		SaveGameOptions (i);
}

/***************************************************************************
    Internal functions
 ***************************************************************************/

static int initialize_rc_winui_config (void)
{
	static char unknown[3];
	static struct rc_option flag_opts[] =
	{
		{ "%s_favorite", NULL, rc_bool, NULL, "0", 0, 0, NULL, "Favorite game" },
		{ "%s_playcount", NULL, rc_int, NULL, "0", 0, (UINT32)-1, NULL, "play counts" },
		{ "%s_has_roms", NULL, rc_int, NULL, unknown, 0, 2, NULL, "has roms" },
		{ "%s_has_samples", NULL, rc_int, NULL, unknown, 0, 2, NULL, "has samples" },
		{ NULL,	NULL, rc_end, NULL, NULL, 0, 0,	NULL, NULL }
	};
	struct rc_option *rc;
	int i, j;

#define REGIST_GAME_OPT(n, item)			\
	rc[n].name = buf;				\
	rc[n].dest = &game[i].item;			\
	buf += strlen (buf) + 1;

#define NUM_FLAG_OPTS	(sizeof flag_opts / sizeof *flag_opts)

	sprintf (unknown, "%d", UNKNOWN);

	rc_winui = (struct rc_struct **)malloc(sizeof (struct rc_struct *) * (num_games + 1));
	if (!rc_winui)
		exit (0);

	if (!(rc_winui[num_games] = rc_create()))
		exit (0);

	if (rc_register(rc_winui[num_games], rc_winui_opts))
		exit(0);

	// Setup default font
	strcpy (settings.list_font.lfFaceName, _(settings.list_font.lfFaceName));
	switch (GetOEMCP())
	{
		case 932:
			settings.list_font.lfCharSet        = SHIFTJIS_CHARSET;
			settings.list_font.lfHeight         = -12;
			break;
		case 936:
			settings.list_font.lfCharSet        = GB2312_CHARSET;
			settings.list_font.lfHeight         = -12;
			break;
		case 950:
			settings.list_font.lfCharSet        = CHINESEBIG5_CHARSET;
			settings.list_font.lfHeight         = -12;
			break;
		default:
			settings.list_font.lfCharSet        = ANSI_CHARSET;
	}

	for (i = 0; i < num_games; i++) {
		char work[NUM_FLAG_OPTS * 32];
		char *buf;
		size_t name_size;

		if (!(rc_winui[i] = rc_create()))
			exit (0);

		name_size = 0;

		for (j = 0; j < NUM_FLAG_OPTS; j++) {
			if (flag_opts[j].name == NULL)
				break;
			sprintf (work + name_size, flag_opts[j].name, drivers[i]->name);
			name_size += strlen (work + name_size) + 1;
		}

		buf = malloc(sizeof flag_opts + name_size);
		if (!buf)
			exit (0);

		rc = (struct rc_option *)buf;
		buf += sizeof flag_opts;

		memcpy (buf, work, name_size);

		for (j = 0; j < NUM_FLAG_OPTS; j++)
			rc[j] = flag_opts[j];

		REGIST_GAME_OPT (0, is_favorite);
		REGIST_GAME_OPT (1, play_count);
		REGIST_GAME_OPT (2, has_roms);
		REGIST_GAME_OPT (3, has_samples);

		if (rc_register(rc_winui[i], rc))
			exit (0);
	}

	return 0;
}

static int rc_load_winui_config (void)
{
	char filename[512];
	char buf[512];
	FILE *fp;
	int i = num_games;
	int line = 0;

	sprintf (filename, "%s.ini", MAME32NAME);

	if (!(fp = fopen(filename, "r")))
		return 0;

    strcpy (rc_dummy_args.save_version, "(unknown)");

	while(fgets(buf, sizeof buf, fp))
	{
		struct rc_option *option = NULL;
		char *name, *tmp, *arg = NULL;
		int last;

		line ++;

		/* get option name */
		if(!(name = strtok(buf, " \t\r\n")))
			continue;
		if(name[0] == '#')
			continue;

		/* get complete rest of line */
		arg = strtok(NULL, "\r\n");

		/* ignore white space */
		for (; (*arg == '\t' || *arg == ' '); arg++) {}

		/* deal with quotations */
		if (arg[0] == '"')
			arg = strtok (arg, "\"");
		else if (arg[0] == '\'')
			arg = strtok (arg, "'");
		else
			arg = strtok (arg, " \t\r\n");

		if (!arg)
		{
			fprintf(stderr,
				"error: %s requires an argument, on line %d of file: %s\n",
				name, line, filename);
				continue;
		}

		last = i;
		for (; i <= num_games; i++)
			if ((option = rc_get_option(rc_winui[i], name)) != NULL)
				break;

		if (!option)
			for (i = 0; i < last; i++)
				if ((option = rc_get_option(rc_winui[i], name)) != NULL)
					break;

		if (!option)
		{
			fprintf(stderr, "error: unknown option %s, on line %d of file: %s\n",
				name, line, filename);
			i = last;
			continue;
		}
		else if ( (tmp = strtok(NULL, " \t\r\n")) && (tmp[0] != '#') )
		{
			fprintf(stderr,
				"error: trailing garbage: \"%s\" on line: %d of file: %s\n",
				tmp, line, filename);
		}
		else if (!rc_set_option3(option, arg, 1))
			continue;

	         fprintf(stderr, "   ignoring line\n");
	}

	fclose (fp);

	return 0;
}

static int rc_save_winui_config (void)
{
	char buffer[128];
	FILE *fp;
	int i;

	sprintf (buffer, "%s.ini", MAME32NAME);

	if (!(fp = fopen(buffer, "w")))
		return 0;

	strcpy (rc_dummy_args.save_version, GetVersionString ());

	ColumnOrderEncodeString ();
	ColumnShownEncodeString ();
	ColumnEncodeWidths ();
	SplitterEncodeString ();
	ListEncodeString ();
	FontEncodeString ();

	rc_write (rc_winui[num_games], fp, buffer);

	rc_write_folder_flags (fp);

	for (i = 0; i < num_games; i++) {
		if ((game[i].has_roms == UNKNOWN) && (game[i].has_samples == UNKNOWN))
			if ((game[i].is_favorite == FALSE) && (game[i].play_count == 0))
				continue;
/*
		UseLangList()?
		rc_write (rc_winui[i], fp, D_("lst",drivers[i]->description)):
*/
		rc_write (rc_winui[i], fp, drivers[i]->description);
	}

	fclose (fp);

	return 0;
}

static int rc_write_folder_flags (FILE *fp)
{
	int i;

	fprintf (fp, "### %s ###\n", _("WINUI specific folder flag options"));

	for (i = 0; i < num_folders; i++)
		if (folder_flags[i].name[0] != '\0')
		{
			fprintf (fp, "%-21s   ", FOLDERFLAG_OPT);
			fprintf (fp, "\"%s,%ld\"\n", folder_flags[i].name, folder_flags[i].flags);
		}
	fprintf (fp, "\n");

	return 0;
}

static int rc_load_config (int game_index)
{
	struct rc_struct *rc;
	char buffer[128];
	FILE *fp;
	int retval;

	if (game_index != -1) {
		gOpts.use_default = TRUE;
		sprintf (buffer, "%s/%s.ini", GetIniDir(), drivers[game_index]->name);
		rc = rc_game;
	}
	else {
		sprintf (buffer, "%s.ini", MAMENAME);
		rc = rc_core;
	}

	free_if_allocated (&rc_dummy_args.playbackname);
	free_if_allocated (&rc_dummy_args.recordname);

	if (!(fp = fopen(buffer, "r")))
		return 0;

	gOpts.use_default = FALSE;

	retval = rc_read (rc, fp, buffer, 1, 1);

	fclose (fp);

	free_if_allocated (&rc_dummy_args.playbackname);
	free_if_allocated (&rc_dummy_args.recordname);

	return retval;
}

static int rc_save_config (int game_index)
{
	struct rc_struct *rc;
	char buffer[128];
	FILE *fp;
	int retval;

	if (game_index != -1) {
		sprintf (buffer, "%s/%s.ini", GetIniDir(), drivers[game_index]->name);
		rc = rc_game;

		if (gOpts.use_default == TRUE) {
			unlink (buffer);
			return 0;
		}
	}
	else {
		sprintf (buffer, "%s.ini", MAMENAME);
		rc = rc_core;
	}

	if (!strcmp (gOpts.resolution, "0x0x0"))
		strcpy (gOpts.resolution, "auto");

	if (!strcmp (gOpts.debugres, "0x0x0"))
		strcpy (gOpts.debugres, "auto");

	if (!(fp = fopen(buffer, "w")))
		return -1;

	if (game_index != -1)
/*
		UseLangList()?
		strcpy (buffer, D_("lst",drivers[game_index]->description)):
*/
		strcpy (buffer, drivers[game_index]->description);

	retval = rc_write (rc, fp, buffer);

	fclose (fp);

	return retval;
}

static void backup_settings (settings_type *p)
{
	p->language =		strdup (p->language);
	p->romdirs =		strdup (p->romdirs);
	p->sampledirs =		strdup (p->sampledirs);
	p->cfgdir =		strdup (p->cfgdir);
	p->hidir =		strdup (p->hidir);
	p->inpdir =		strdup (p->inpdir);
	p->imgdir =		strdup (p->imgdir);
	p->statedir =		strdup (p->statedir);
	p->artdir =		strdup (p->artdir);
	p->memcarddir =		strdup (p->memcarddir);
	p->flyerdir =		strdup (p->flyerdir);
	p->cabinetdir =		strdup (p->cabinetdir);
	p->marqueedir =		strdup (p->marqueedir);
#ifdef PLUS
	p->cpaneldir =		strdup (p->cpaneldir);
	p->titledir =		strdup (p->titledir);
#endif /* PLUS */
	p->nvramdir =		strdup (p->nvramdir);
	p->diffdir =		strdup (p->diffdir);
	p->inidir =		strdup (p->inidir);
	p->iconsdir =		strdup (p->iconsdir);
#ifdef EXTRA_FOLDER
	p->folderdir  =		strdup (p->folderdir);
#endif /* EXTRA_FOLDER */
	p->cheatfile =		strdup (p->cheatfile);
	p->history_filename =	strdup (p->history_filename);
	p->mameinfo_filename =	strdup (p->mameinfo_filename);
	p->ctrlrdir =		strdup (p->ctrlrdir);
}

static void free_settings (settings_type *p)
{
	free_if_allocated (&p->language);
	free_if_allocated (&p->romdirs);
	free_if_allocated (&p->sampledirs);
	free_if_allocated (&p->cfgdir);
	free_if_allocated (&p->hidir);
	free_if_allocated (&p->inpdir);
	free_if_allocated (&p->imgdir);
	free_if_allocated (&p->statedir);
	free_if_allocated (&p->artdir);
	free_if_allocated (&p->memcarddir);
	free_if_allocated (&p->flyerdir);
	free_if_allocated (&p->cabinetdir);
	free_if_allocated (&p->marqueedir);
#ifdef PLUS
	free_if_allocated (&p->cpaneldir);
	free_if_allocated (&p->titledir);
#endif /* PLUS */
	free_if_allocated (&p->nvramdir);
	free_if_allocated (&p->diffdir);
	free_if_allocated (&p->inidir);
	free_if_allocated (&p->iconsdir);
#ifdef EXTRA_FOLDER
	free_if_allocated (&p->folderdir);
#endif /* EXTRA_FOLDER */
	free_if_allocated (&p->cheatfile);
	free_if_allocated (&p->history_filename);
	free_if_allocated (&p->mameinfo_filename);
	free_if_allocated (&p->ctrlrdir);
}

static void free_if_allocated (char **p)
{
	if (*p)
		free (*p);
	*p = NULL;
}

static int ColumnDecodeString (struct rc_option *option, const char *arg, int *data)
{
	int  i;
	char *s, *p;
	int temp[COLUMN_MAX];
	char buf[80];

	if (arg == NULL)
		return -1;

	strcpy (buf, arg);
	p = buf;

	for (i = 0; p && i < COLUMN_MAX; i++)
	{
		s = p;
		
		if ((p = strchr (s, ',')) != NULL && *p == ',')
		{
			*p = '\0';
			p++;
		}
		temp[i] = atoi (s);
	}

	if ((i != COLUMN_MAX) || (p))
	{
		fprintf(stderr, "error invalid value for %s: %s\n", option->name, arg);
		return -1;
	}

	for (i = 0; i < COLUMN_MAX; i++)
		data[i] = temp[i];

	return 0;
}

static char *ColumnEncodeString (int *data)
{
	int  i;
	static char str[80];

	sprintf (str, "%d", data[0]);
	
	for (i = 1; i < COLUMN_MAX; i++)
		sprintf (str + strlen (str), ",%d", data[i]);

	return str;
}

static int ColumnOrderDecodeString (struct rc_option *option, const char *arg, int priority)
{
	return ColumnDecodeString (option, arg, settings.column_order);
}

static void ColumnOrderEncodeString (void)
{
	strcpy (rc_dummy_args.column_order, ColumnEncodeString (settings.column_order));
}

static int ColumnShownDecodeString (struct rc_option *option, const char *arg, int priority)
{
	return ColumnDecodeString (option, arg, settings.column_shown);
}

static void  ColumnShownEncodeString (void)
{
	strcpy (rc_dummy_args.column_shown, ColumnEncodeString (settings.column_shown));
}

static int ColumnDecodeWidths (struct rc_option *option, const char *arg, int priority)
{
//	if (settings.view == VIEW_REPORT)
	return ColumnDecodeString (option, arg, settings.column_width);
}

static void ColumnEncodeWidths (void)
{
	strcpy (rc_dummy_args.column_width, ColumnEncodeString (settings.column_width));
}

static int SplitterDecodeString (struct rc_option *option, const char *arg, int priority)
{
	int  i;
	char *s, *p;
	int temp[SPLITTER_MAX];
	char buf[80];

	if (arg == NULL)
		return -1;

	strcpy (buf, arg);
	p = buf;

	for (i = 0; p && i < SPLITTER_MAX; i++)
	{
		s = p;
		
		if ((p = strchr (s, ',')) != NULL && *p == ',')
		{
			*p = '\0';
			p++;
		}
		temp[i] = atoi (s);
	}

	if ((i != SPLITTER_MAX) || (p))
	{
		fprintf(stderr, "error invalid value for %s: %s\n", option->name, arg);
		return -1;
	}

	for (i = 0; i < SPLITTER_MAX; i++)
		settings.splitter[i] = temp[i];

	return 0;
}

static void SplitterEncodeString (void)
{
	int  i;
	char str[80];

	sprintf (str, "%d", settings.splitter[0]);
	
	for (i = 1; i < SPLITTER_MAX; i++)
		sprintf (str + strlen (str), ",%d", settings.splitter[i]);

	strcpy (rc_dummy_args.splitter, str);
}

static int ListDecodeString (struct rc_option *option, const char *arg, int priority)
{
	int i;

	settings.view = VIEW_REPORT;

	for (i = VIEW_LARGE_ICONS; i < VIEW_MAX; i++)
	{
		if (strcmp (arg, view_modes[i]) == 0)
		{
			settings.view = i;
			return 0;
		}
	}

	return -1;
}

static void ListEncodeString (void)
{
	strcpy (rc_dummy_args.view, view_modes[settings.view]);
}

/* Parse the given comma-delimited string into a LOGFONT structure */
static int FontDecodeString (struct rc_option *option, const char *arg, int priority)
{
	LOGFONT* f = &settings.list_font;

	sscanf (arg, "%li,%li,%li,%li,%li,%i,%i,%i,%i,%i,%i,%i,%i",
		   &f->lfHeight,
		   &f->lfWidth,
		   &f->lfEscapement,
		   &f->lfOrientation,
		   &f->lfWeight,
		   (int*)&f->lfItalic,
		   (int*)&f->lfUnderline,
		   (int*)&f->lfStrikeOut,
		   (int*)&f->lfCharSet,
		   (int*)&f->lfOutPrecision,
		   (int*)&f->lfClipPrecision,
		   (int*)&f->lfQuality,
		   (int*)&f->lfPitchAndFamily);

	return 0;
}

/* Encode the given LOGFONT structure into a comma-delimited string */
static void FontEncodeString (void)
{
	LOGFONT* f = &settings.list_font;

	sprintf (rc_dummy_args.list_font, "%li,%li,%li,%li,%li,%i,%i,%i,%i,%i,%i,%i,%i",
			f->lfHeight,
			f->lfWidth,
			f->lfEscapement,
			f->lfOrientation,
			f->lfWeight,
			f->lfItalic,
			f->lfUnderline,
			f->lfStrikeOut,
			f->lfCharSet,
			f->lfOutPrecision,
			f->lfClipPrecision,
			f->lfQuality,
			f->lfPitchAndFamily);
}

static int FontDecodeColor (struct rc_option *option, const char *arg, int priority)
{
	SetListFontColor (settings.list_font_color);

	return 0;
}

static int FolderFlagDecodeString (struct rc_option *option, const char *arg, int priority)
{
	char str[80], *p;

	if (arg == NULL)
	return 0;

	strcpy (str, arg);
	p = strrchr (str, ',');
	if (p == NULL)
	{
		fprintf(stderr, "error invalid value for %s: %s\n", option->name, arg);
		return -1;
	}

	*p++ = '\0';

	SaveFolderFlags (str, atoi (p));

	return 0;
}

static void SaveFolderFlags (char *folderName, DWORD dwFlags)
{
	int i;

	for (i = 0; i < num_folders; i++)
		if (!strcmp (folder_flags[i].name, folderName))
		{
			if (dwFlags == 0)
				folder_flags[i].name[0] = '\0';
			else
				folder_flags[i].flags = dwFlags;
			return;
		}

	if (dwFlags == 0)
		return;

	for (i = 0; i < num_folders; i++)
		if (folder_flags[i].name[0] == '\0')
			break;

	if (i == num_folders)
	{
		struct folder_flags_type *tmp;

		tmp = realloc (folder_flags, (num_folders + ALLOC_FOLDERFLAG) * sizeof (struct folder_flags_type));
		if (!tmp)
		{
			fprintf(stderr, "error: realloc failed in SaveFolderFlags\n");
			return;
		}

		folder_flags = tmp;
		memset (tmp + num_folders, 0, ALLOC_FOLDERFLAG * sizeof (struct folder_flags_type));
		num_folders += ALLOC_FOLDERFLAG;
	}

	strcpy (folder_flags[i].name, folderName);
	folder_flags[i].flags = dwFlags;
}

static void ValidateDefaultOptions (void)
{
	global.use_default	= FALSE;
	global.play_count	= 0;
	global.has_roms		= UNKNOWN;
	global.has_samples	= UNKNOWN;
	global.is_favorite	= FALSE;
}

static void LoadGameOptions (int game_index)
{
	assert (0 <= game_index && game_index < num_games);

	gOpts = game[game_index];

	rc_load_config (game_index);

	game[game_index] = gOpts;
}

static void LoadDefaultOptions (void)
{
	gOpts = backup.global;

	rc_load_config (-1);

	global = gOpts;

	ValidateDefaultOptions ();
}

static void LoadOptions (void)
{
	int   i;

	LoadDefaultOptions ();
	rc_load_winui_config ();

	if (!bResetGUI && settings.version_check)
	{
		char *current_version;

		current_version = GetVersionString ();

		if (*rc_dummy_args.save_version && strcmp (rc_dummy_args.save_version, current_version) != 0)
		{
			char msg[400];
			sprintf (msg,_(oldInfoMsg), rc_dummy_args.save_version, current_version);
			if (MessageBox (0, msg, _("Version Mismatch"), MB_YESNO | MB_ICONQUESTION) == IDYES)
				bResetGUI = TRUE;
		}
	}

	if (bResetGUI)
	{
		free_settings (&settings);
		settings = backup.settings;
	}
	else
		free_settings (&backup.settings);

	bResetGUI = FALSE;

	for (i = 0 ; i < num_games; i++)
		LoadGameOptions (i);
}

static void SaveGlobalOptions (void)
{
	rc_save_winui_config ();
	SaveDefaultOptions ();
}

/* End of options.c */
